/*
	RCSFWD.c
*/

#include <windows.h>
#include <stdio.h>

int main(int argc, char *argv[]);
int service_install(char *service_name);
int service_uninstall(char *service_name);
void service_reportstatus(DWORD curstate, DWORD exitcode, DWORD waithint);
VOID WINAPI service_ctrl(DWORD ctrl);
DWORD WINAPI service_thread(LPVOID param);
static void __stdcall service_main(DWORD argc, LPTSTR *argv);

HANDLE svc_h = NULL;
SERVICE_STATUS svc_ss; 
SERVICE_STATUS_HANDLE svc_ssh; 
HANDLE pid = NULL;

int main(int argc, char *argv[])
{
	SERVICE_TABLE_ENTRY dispatch_table[] = {
        { "", service_main },
        { NULL, NULL }
    };

    if(argc > 1) {
		if(!strcmp(argv[1], "--install")) {
			return service_install("RCSFWD");
		} else if(!strcmp(argv[1], "--uninstall")) {
			return service_uninstall("RCSFWD");
		} else if(!strcmp(argv[1], "--test")) {
			return service_thread(0);
		} else {
			printf("usage: RCSFWD [--install|--uninstall|--test]\n");
			return 0;
		}
	}

    if(!StartServiceCtrlDispatcher(dispatch_table)) {
		return -1;
    }

	return 0;
}

int service_install(char *service_name)
{
    char service_path[MAX_PATH];
	SC_HANDLE schService, schSCManager;
	
    if(GetModuleFileName(NULL, service_path, sizeof(service_path)) == 0) {
		printf("error: GetModuleFileName (%d)\n", GetLastError());
        return -1;
    }

	if(!(schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE))) {
		printf("error: OpenSCManager (%d)\n", GetLastError());
		return -1;
	}

	if(!(schService = CreateService(schSCManager,
									service_name,
									service_name,
									SERVICE_ALL_ACCESS,
									SERVICE_WIN32_OWN_PROCESS,
									SERVICE_AUTO_START,
									SERVICE_ERROR_NORMAL,
									service_path,
									NULL,
									NULL,
									NULL,
									NULL,
									NULL)
	)) {
		printf("error: CreateService (%d)\n", GetLastError());
		return -1;
	}

	CloseServiceHandle(schService);
	CloseServiceHandle(schSCManager);

	printf("Service installed successfully\n"); 

    return 0;
}

int service_uninstall(char *service_name)
{
	SC_HANDLE schService, schSCManager;

	if(!(schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_CONNECT))) {
		printf("error: OpenSCManager (%d)\n", GetLastError());
		return -1;
	}

	if(!(schService = OpenService(schSCManager, service_name, DELETE))) {
		printf("error: OpenService (%d)\n", GetLastError());
	   return -1;
	}

	if (DeleteService(schService) == 0) {
		printf("error: DeleteService (%d)\n", GetLastError());
		return -1;
	}

	CloseServiceHandle(schService);
	CloseServiceHandle(schSCManager);

	printf("Service uninstalled successfully\n"); 
	
    return 0;
}

void service_reportstatus(DWORD curstate, DWORD exitcode, DWORD waithint)
{
    static DWORD checkpoint = 1;

    svc_ss.dwCurrentState = curstate;
    svc_ss.dwWin32ExitCode = exitcode;
    svc_ss.dwWaitHint = waithint;

    if(curstate == SERVICE_START_PENDING) {
        svc_ss.dwControlsAccepted = 0;
    } else {
		svc_ss.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	}
	
    if((curstate == SERVICE_RUNNING) || (curstate == SERVICE_STOPPED)) {
        svc_ss.dwCheckPoint = 0;
    } else {
		svc_ss.dwCheckPoint = checkpoint++;
	}

    SetServiceStatus(svc_ssh, &svc_ss);
	
	return;
}

VOID WINAPI service_ctrl(DWORD ctrl)
{
	switch(ctrl) {  
		case SERVICE_CONTROL_STOP: 
			service_reportstatus(SERVICE_STOP_PENDING, NO_ERROR, 0);
			if(pid) WaitForSingleObject(pid, INFINITE);
			SetEvent(svc_h);
			break; 
		default: 
			service_reportstatus(svc_ss.dwCurrentState, NO_ERROR, 0);
			break;
	} 

	return;
}

DWORD WINAPI service_thread(LPVOID param)
{
	char service_path[MAX_PATH], *bs;
    STARTUPINFO si;
    PROCESS_INFORMATION pi;

    if(GetModuleFileName(NULL, service_path, sizeof(service_path)) == 0) {
		printf("error: GetModuleFileName (%d)\n", GetLastError());
        return -1;
    }

	if((bs = strrchr(service_path, '\\')) != NULL) bs[0] = '\0';
	if((bs = strrchr(service_path, '\\')) != NULL) bs[0] = '\0';
	
	if(!SetCurrentDirectory(service_path)) {
		printf("error: SetCurrentDirectory (%d)\n", GetLastError());
        return -1;
    }

    while(1) {
		ZeroMemory(&si, sizeof(si));
		si.cb = sizeof(si);
		ZeroMemory(&pi, sizeof(pi));

		if(!CreateProcess(NULL,
						  "php bin\\RCSFWD.php",
						  NULL,
						  NULL,
						  FALSE,
						  0,
						  NULL,
						  NULL,
						  &si,
						  &pi)
		) {
			printf("error: CreateProcess (%d)\n", GetLastError());
			return -1;
		}

		pid = pi.hProcess;

		WaitForSingleObject(pid, INFINITE);

		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}

	return -1;
}

static void __stdcall service_main(DWORD argc, LPTSTR *argv)
{
	DWORD thread_id;
	HANDLE thread_h = NULL;

    if(!(svc_ssh = RegisterServiceCtrlHandler(argv[0], service_ctrl))) {
		printf("error: RegisterServiceCtrlHandler (%d)\n", GetLastError());
        return; 
    } 

    svc_ss.dwServiceType = SERVICE_WIN32_OWN_PROCESS; 
    svc_ss.dwServiceSpecificExitCode = 0;    

    service_reportstatus(SERVICE_START_PENDING, NO_ERROR, 3000);

    if((svc_h = CreateEvent(NULL, TRUE, FALSE, NULL)) == NULL) {
		printf("error: CreateEvent (%d)\n", GetLastError());
        service_reportstatus(SERVICE_STOPPED, NO_ERROR, 0);
        return;
    }

	if((thread_h = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)service_thread, NULL, 0, &thread_id)) == NULL) {
		printf("error: CreateThread (%d)\n", GetLastError());
		return;
	}

    service_reportstatus(SERVICE_RUNNING, NO_ERROR, 0);
    WaitForSingleObject(svc_h, INFINITE);
    service_reportstatus(SERVICE_STOPPED, NO_ERROR, 0);

	return;
}
